/**
 * Copyright (C) (2010-2018) Vadim Biktashev, Irina Biktasheva et al. 
 * (see ../AUTHORS for the full list of contributors)
 *
 * This file is part of Beatbox.
 *
 * Beatbox is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Beatbox is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Beatbox.  If not, see <http://www.gnu.org/licenses/>.
 */

/* Wrapper around ezstep & ezstep3d from D. Barkley's ezspiral and ezscroll.  */
/* For now, periodic directions are not allowed.                              */
/* To do: MPI                                                                 */

#include <math.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "system.h"
#include "beatbox.h"
#include "device.h"
#include "state.h"
#include "bikt.h"
#include "qpp.h"

typedef struct {
  /* Front-end, specified in input file */
  #define _(Type,Name,Dflt,miN,maX) Type Name;
  #include "ezstep.h"
  #undef _
  /* Back-end, ezspiral/ezscroll computed constants */
  real dt;
  real one_o_a;
  real b_o_a;
  real dt_o_eps;
  real dt_o_2eps;
  real dt_o_wh2;
  real dtDv_o_wh2;
  int v_diff_on;		/* if 1 then v-field diffuses			*/
  /* Auxiliary variables */
  int s1;
  int s2;
  int v0;
  int first_time;
} STR;

/* True constants */
static const real zero       = 0.;
static const real one        = 1.;
static const real two        = 2.;
static const real four       = 4.;
static const real six        = 6.;
static const real twenty     = 20.;
static const real twentyfour = 24.;

/* EZ/Beatbox matching macros */
#define NX (xmax-2)
#define NY (ymax-2)
#define NZ (zmax-2)
#define I_INC (DX)
#define J_INC (DY)
#define K_INC (DZ)

static void Step_ini_2D (STR *S);
static void Step_ini_3D (STR *S);
static void Impose_boundary_conditions_2D (STR *S, real *w, real *sigma_w);
static void Impose_boundary_conditions_3D (STR *S, real *w, real *sigma_w);

/****************/
RUN_HEAD(ezstep)
{
  DEVICE_CONST(real,delta);
  DEVICE_CONST(int,v_diff_on);
  DEVICE_CONST(int,explicit);
  DEVICE_CONST(int,split);
  DEVICE_CONST(int,manypoint);
  DEVICE_CONST(real,dt);
  DEVICE_CONST(real,one_o_a);
  DEVICE_CONST(real,b_o_a);
  DEVICE_CONST(real,dt_o_eps);
  DEVICE_CONST(real,dt_o_2eps);
  DEVICE_CONST(real,dt_o_wh2);
  DEVICE_CONST(real,dtDv_o_wh2);
  DEVICE_VAR(int,s1);
  DEVICE_VAR(int,s2);
  DEVICE_VAR(int,first_time);
  real u_thresh;
  register int i, j, k;
  register int index, index_1, index_2;
  real *u       = &(New[ind(0,0,0,s.v0+0)]);
  real *v       = &(New[ind(0,0,0,s.v0+1)]);
  real *sigma_u = &(New[ind(0,0,0,s.v0+2)]);
  real *sigma_v = &(New[ind(0,0,0,s.v0+4)]);

  if (*first_time) {
    /* Initialize arrays for timestepping */
    switch (dim) {
    case 2: Step_ini_2D(S); break;
    case 3: Step_ini_3D(S); break;
    default: EXPECTED_ERROR("ezstep is not implemented for %d dimensions\n",dim);
    }
    *first_time=0;
  }
  
  /* Interchange s1 and s2 */
  (*s1) = 1-(*s1);
  (*s2) = 1-(*s2);

  /* A 5-fold #include chain that converts integer flags to CPP macros */
  if (v_diff_on) {
    #define V_DIFF_ON 1
    #include "ezstep1.h"
    #undef V_DIFF_ON
  } else {
    #define V_DIFF_ON 0
    #include "ezstep1.h"
    #undef V_DIFF_ON
  } /* if diff_on else */
  
} RUN_TAIL(ezstep);

DESTROY_HEAD(ezstep)
DESTROY_TAIL(ezstep)

CREATE_HEAD(ezstep)
{
  DEVICE_REQUIRES_SYNC
  real WEIGHT, STABILITY_LIMIT;

  /* Front-end parameters are read from the script */
  #define acceptreal(b,c,d,e) if (!acceptr(#b"=",&(S->b),c,d,e,w)) return(0); real b=S->b
  #define acceptint(b,c,d,e) if (!accepti(#b"=",&(S->b),c,d,e,w)) return(0); int b=S->b 
  #define _(Type,Name,Dflt,miN,maX) accept##Type(Name,Dflt,miN,maX);
  #include "ezstep.h"
  #undef _
  #undef acceptint
  #undef acceptreal

  /* Some checks on the front-end parameters */
  if (S->manypoint) {
    WEIGHT=6.;
    STABILITY_LIMIT=3.*hx*hx/8.;
  } else {
    WEIGHT=1.;
    switch (dim) {
    case 2: STABILITY_LIMIT=hx*hx/4.; break;
    case 3: STABILITY_LIMIT=hx*hx/6.; break;
    default: EXPECTED_ERROR("%s not implemented for %d dimensions\n",dev->n,dim);
    }
  }
  ASSERT(ht < STABILITY_LIMIT);
  ASSERT(a != 0);
  ASSERT(eps != 0);
  S->v_diff_on = (0.0 != S->Dv);
  if (S->v_diff_on) {
    if (dev->s.v1 - dev->s.v0 < 5 )
      EXPECTED_ERROR("%s device with second diffusion requires six contiguous layers; given %d..%d do not satisfy that\n",dev->n,dev->s.v0,dev->s.v1);
  } else {
    if (dev->s.v1 - dev->s.v0 < 3 )
      EXPECTED_ERROR("%s device without second diffusion requires four contiguous layers; given %d..%d do not satisfy that\n",dev->n,dev->s.v0,dev->s.v1);
  }
  
  /* Back-end parameters calculated from the front-end ones */
  #define _(var,expr) real var=expr; S->var=var; printf("%s=%g\n",#var,var);
  _(dt , ht)
  _(one_o_a , 1.0/a)
  _(b_o_a , b/a)
  _(dt_o_eps , dt/eps)
  _(dt_o_2eps  , dt_o_eps/2.) 
  _(dt_o_wh2 , D*dt/(WEIGHT*hx*hx))
  _(dtDv_o_wh2 , Dv/D*dt_o_wh2)
  _(v0 , dev->s.v0)
  #undef _

  /* Initialize arrays for timestepping at the first run, not now */
  S->first_time=1;
}
CREATE_TAIL(ezstep,1)
/* ========================================================================= */

static void Step_ini_2D (STR *S)
{
  DEVICE_CONST(int,manypoint);
  DEVICE_CONST(int,v_diff_on);
  DEVICE_VAR(int,s1);
  DEVICE_VAR(int,s2);
  DEVICE_CONST(int,v0);
  int i, j;
  int index, index_1, index_2;
  real *u=&(New[ind(0,0,0,v0)]);
  real *v=&(New[ind(0,0,0,v0+1)]);
  real *sigma_u=&(New[ind(0,0,0,v0+2)]);
  real *sigma_v=v_diff_on?(&(New[ind(0,0,0,v0+4)])):NULL;

  /* Set initial s1 and s2 */
  (*s1) = 0;
  (*s2) = 1;

  /* Initialize spatial sums. */
  switch (2*v_diff_on+1*manypoint) {
  #define ZERO_USED_SUMS(index_1)  sigma_u[index_1] = zero;
  case 2*0+1*0:
    for(j=1;j<=NY;j++) {
      for(i=1;i<=NX;i++) {
	index = ind(i,j,0,0);
	index_1 = index + (*s1)*DV;
	index_2 = index + (*s2)*DV;
	ADD_TO_U_SUM5(index,index_2);
	ZERO_USED_SUMS(index_1);
      }
    }
    break;
  case 2*0+1*1:
    for(j=1;j<=NY;j++) {
      for(i=1;i<=NX;i++) {
	index = ind(i,j,0,0);
	index_1 = index + (*s1)*DV;
	index_2 = index + (*s2)*DV;
	ADD_TO_U_SUM9(index,index_2);
	ZERO_USED_SUMS(index_1);
      }
    }
    break;
  #undef ZERO_USED_SUMS
  #define ZERO_USED_SUMS(index_1)  sigma_u[index_1] = sigma_v[index_1] = zero;
  case 2*1+1*0:
    for(j=1;j<=NY;j++) {
      for(i=1;i<=NX;i++) {
	index = ind(i,j,0,0);
	index_1 = index + (*s1)*DV;
	index_2 = index + (*s2)*DV;
	ADD_TO_U_SUM5(index,index_2);
	ADD_TO_V_SUM5(index,index_2);
	ZERO_USED_SUMS(index_1);
      }
    }
    break;
  case 2*1+1*1:
    for(j=1;j<=NY;j++) {
      for(i=1;i<=NX;i++) {
	index = ind(i,j,0,0);
	index_1 = index + (*s1)*DV;
	index_2 = index + (*s2)*DV;
	ADD_TO_U_SUM9(index,index_2);
	ADD_TO_V_SUM9(index,index_2);
	ZERO_USED_SUMS(index_1);
      }
    }
    break;
  #undef ZERO_USED_SUMS
  }/* switch v_diff_on and manypoint */
  Impose_boundary_conditions_2D(S, u, sigma_u);
  if (v_diff_on) 
    Impose_boundary_conditions_2D(S, v, sigma_v);
} /* Step_ini_2D */
/* ========================================================================= */

static void Step_ini_3D (STR *S)
{
  DEVICE_CONST(int,manypoint);
  DEVICE_CONST(int,v_diff_on);
  DEVICE_VAR(int,s1);
  DEVICE_VAR(int,s2);
  DEVICE_CONST(int,v0);
  int i, j, k;
  int index, index_1, index_2;
  real *u=&(New[ind(0,0,0,v0)]);
  real *v=&(New[ind(0,0,0,v0+1)]);
  real *sigma_u=&(New[ind(0,0,0,v0+2)]);
  real *sigma_v=v_diff_on?(&(New[ind(0,0,0,v0+4)])):NULL;

  /* Set initial (*s1) and (*s2) */
  (*s1) = 0;
  (*s2) = 1;

  /* Initialize spatial sums. */
  switch (2*v_diff_on+1*manypoint) {
  #define ZERO_USED_SUMS(index_1)  sigma_u[index_1] = zero;
  case 2*0+1*0:
    for(k=1;k<=NZ;k++) {
      for(j=1;j<=NY;j++) {
	for(i=1;i<=NX;i++) {
	  index = ind(i,j,k,0);
	  index_1 = index + (*s1)*DV;
	  index_2 = index + (*s2)*DV;
	  ADD_TO_U_SUM7(index,index_2);
	  ZERO_USED_SUMS(index_1);
	}
      }
    }
    break;
  case 2*0+1*1:
    for(k=1;k<=NZ;k++) {
      for(j=1;j<=NY;j++) {
	for(i=1;i<=NX;i++) {
	  index = ind(i,j,k,0);
	  index_1 = index + (*s1)*DV;
	  index_2 = index + (*s2)*DV;
	  ADD_TO_U_SUM19(index,index_2);
	  ZERO_USED_SUMS(index_1);
	}
      }
    }
    break;
  #undef ZERO_USED_SUMS
  #define ZERO_USED_SUMS(index_1)  sigma_u[index_1] = sigma_v[index_1] = zero;
  case 2*1+1*0:
    for(k=1;k<=NZ;k++) {
      for(j=1;j<=NY;j++) {
	for(i=1;i<=NX;i++) {
	  index = ind(i,j,k,0);
	  index_1 = index + (*s1)*DV;
	  index_2 = index + (*s2)*DV;
	  ADD_TO_U_SUM7(index,index_2);
	  ADD_TO_V_SUM7(index,index_2);
	  ZERO_USED_SUMS(index_1);
	}
      }
    }
    break;
  case 2*1+1*1:
    for(k=1;k<=NZ;k++) {
      for(j=1;j<=NY;j++) {
	for(i=1;i<=NX;i++) {
	  index = ind(i,j,k,0);
	  index_1 = index + (*s1)*DV;
	  index_2 = index + (*s2)*DV;
	  ADD_TO_U_SUM19(index,index_2);
	  ADD_TO_V_SUM19(index,index_2);
	  ZERO_USED_SUMS(index_1);
	}
      }
    }
    break;
  #undef ZERO_USED_SUMS
  }/* switch v_diff_on and manypoint */
  Impose_boundary_conditions_3D(S, u, sigma_u);
  if (v_diff_on) 
    Impose_boundary_conditions_3D(S, v, sigma_v);
} /* Step_ini_3D */
/* ========================================================================= */


static void  Impose_boundary_conditions_2D (STR *S, real *w, real *sigma_w)
{
  DEVICE_CONST(int,manypoint);
  DEVICE_CONST(int,s2);
  /* DEVICE_CONST(int,v0); */
  int i,j;
#define W(i,j)          w[ind(i,j,0,0)]
#define Sigma_w(s,i,j)  sigma_w[ind(i,j,0,s)]

  
  /* First the values of u on the "real" part of the mesh are copied to
   * "fictitious" boundary points.  The way points are copied depend on
   * whether Neumann or periodic boundary conditions are being imposed.
   *
   * Note that the indices i, j do not range over the same values in each
   * case (x, y).  This is correct though a bit subtle. */

  /* Set fictitious points in x-direction */
  for(j=1;j<=NY;j++) { 
    W(0   ,j) = W(2   ,j);  
    W(NX+1,j) = W(NX-1,j);  
  }

  /* Set fictitious points in y-direction */
  for(i=0;i<=NX+1;i++) { 
    W(i,0   ) = W(i,2   );  
    W(i,NY+1) = W(i,NY-1);  
  }

  if (manypoint) {
    
    /* -----------------------------
     * Nine-point Laplacian formulas
     * ----------------------------- */
    
    for(i=1;i<=NX;i++) {
      Sigma_w(s2,i  ,1) += four*W(i,0);  
      Sigma_w(s2,i+1,1) +=      W(i,0);  
      Sigma_w(s2,i-1,1) +=      W(i,0); 
      
      Sigma_w(s2,i  ,NY) += four*W(i,NY+1); 
      Sigma_w(s2,i+1,NY) +=      W(i,NY+1);  
      Sigma_w(s2,i-1,NY) +=      W(i,NY+1); 
    } 
    
    for(j=1;j<=NY;j++) {
      Sigma_w(s2,1,j  ) += four*W(0,j); 
      Sigma_w(s2,1,j+1) +=      W(0,j);  
      Sigma_w(s2,1,j-1) +=      W(0,j); 
      
      Sigma_w(s2,NX,j  ) += four*W(NX+1,j); 
      Sigma_w(s2,NX,j+1) +=      W(NX+1,j);  
      Sigma_w(s2,NX,j-1) +=      W(NX+1,j); 
    }
    
    Sigma_w(s2,1 ,1 ) += W(  0 ,  0 );  
    Sigma_w(s2,NX,1 ) += W(NX+1,  0 );  
    Sigma_w(s2,1 ,NY) += W(  0 ,NY+1);  
    Sigma_w(s2,NX,NY) += W(NX+1,NY+1);  

  } else { /* manypoint */

    /* -----------------------------
     * Five-point Laplacian formulas 
     * ----------------------------- */
    
    for(i=1;i<=NX;i++) {
      Sigma_w(s2,i,1 ) += W(i,0   );  
      Sigma_w(s2,i,NY) += W(i,NY+1); 
    }
    
    for(j=1;j<=NY;j++) {
      Sigma_w(s2,1 ,j) += W(0   ,j); 
      Sigma_w(s2,NX,j) += W(NX+1,j); 
    }
  } /* if manypoint else */
#undef W
#undef Sigma_w
} /* Impose_boundary_conditions_2D */
/* ========================================================================= */

static void  Impose_boundary_conditions_3D (STR *S, real *w, real *sigma_w)
{
  DEVICE_CONST(int,manypoint);
  DEVICE_CONST(int,s2);
  /* DEVICE_CONST(int,v0); */
  int i,j,k;
#define W(i,j,k)          w[ind(i,j,k,0)]
#define Sigma_w(s,i,j,k)  sigma_w[ind(i,j,k,s)]


  /* First the values of w on the "real" part of the mesh are copied to
   * "fictitious" boundary points.  The way points are copied depend on
   * whether Neumann or periodic boundary conditions are being imposed.
   *
   * Note that the indices i, j, k do not range over the same values in each
   * case (x, y, z).  This is correct though a bit subtle. */

  /* Set fictitious points in x-direction */
  for(k=1;k<=NZ;k++) { 
    for(j=1;j<=NY;j++) { 
      W(0   ,j,k) = W(2   ,j,k);  
      W(NX+1,j,k) = W(NX-1,j,k);  
    }
  }

  /* Set fictitious points in y-direction */
  for(k=1;k<=NZ;k++) { 
    for(i=0;i<=NX+1;i++) { 
      W(i,0   ,k) = W(i,2   ,k);  
      W(i,NY+1,k) = W(i,NY-1,k);  
    }
  }

  /* Set fictitious points in z-direction */
  for(j=0;j<=NY+1;j++) { 
    for(i=0;i<=NX+1;i++) { 
      W(i,j,0   ) = W(i,j,2   );  
      W(i,j,NZ+1) = W(i,j,NZ-1);  
    }
  }

  if (manypoint) {

/* ---------------------------------
 * Nineteen-point Laplacian formulas
 * --------------------------------- */

    for(j=1;j<=NY;j++) {
      for(i=1;i<=NX;i++) {
	Sigma_w(s2,i  ,j  ,1) += two*W(i,j,0);  
	Sigma_w(s2,i+1,j  ,1) +=     W(i,j,0);  
	Sigma_w(s2,i-1,j  ,1) +=     W(i,j,0); 
	Sigma_w(s2,i  ,j+1,1) +=     W(i,j,0);  
	Sigma_w(s2,i  ,j-1,1) +=     W(i,j,0); 
	
	Sigma_w(s2,i  ,j  ,NZ) += two*W(i,j,NZ+1); 
	Sigma_w(s2,i+1,j  ,NZ) +=     W(i,j,NZ+1);  
	Sigma_w(s2,i-1,j  ,NZ) +=     W(i,j,NZ+1); 
	Sigma_w(s2,i  ,j+1,NZ) +=     W(i,j,NZ+1);  
	Sigma_w(s2,i  ,j-1,NZ) +=     W(i,j,NZ+1); 
      }
    }
    
    for(k=1;k<=NZ;k++) {
      for(i=1;i<=NX;i++) {
	Sigma_w(s2,i  ,1,k  ) += two*W(i,0,k);  
	Sigma_w(s2,i+1,1,k  ) +=     W(i,0,k);  
	Sigma_w(s2,i-1,1,k  ) +=     W(i,0,k); 
	Sigma_w(s2,i  ,1,k+1) +=     W(i,0,k);  
	Sigma_w(s2,i  ,1,k-1) +=     W(i,0,k); 
	
	Sigma_w(s2,i  ,NY,k  ) += two*W(i,NY+1,k); 
	Sigma_w(s2,i+1,NY,k  ) +=     W(i,NY+1,k);  
	Sigma_w(s2,i-1,NY,k  ) +=     W(i,NY+1,k); 
	Sigma_w(s2,i  ,NY,k+1) +=     W(i,NY+1,k);  
	Sigma_w(s2,i  ,NY,k-1) +=     W(i,NY+1,k); 
      }
    } 
    
    for(k=1;k<=NZ;k++) {
      for(j=1;j<=NY;j++) {
	Sigma_w(s2,1,j  ,k  ) += two*W(0,j,k); 
	Sigma_w(s2,1,j+1,k  ) +=     W(0,j,k);  
	Sigma_w(s2,1,j-1,k  ) +=     W(0,j,k); 
	Sigma_w(s2,1,j  ,k+1) +=     W(0,j,k);  
	Sigma_w(s2,1,j  ,k-1) +=     W(0,j,k); 
	
	Sigma_w(s2,NX,j  ,k  ) += two*W(NX+1,j,k); 
	Sigma_w(s2,NX,j+1,k  ) +=     W(NX+1,j,k);  
	Sigma_w(s2,NX,j-1,k  ) +=     W(NX+1,j,k); 
	Sigma_w(s2,NX,j  ,k+1) +=     W(NX+1,j,k);  
	Sigma_w(s2,NX,j  ,k-1) +=     W(NX+1,j,k); 
      }
    }
    
    for(i=1;i<=NX;i++) {
      Sigma_w(s2,i,1 ,1 ) += W(i,  0 ,  0 );  
      Sigma_w(s2,i,NY,1 ) += W(i,NY+1,  0 );  
      Sigma_w(s2,i,1 ,NZ) += W(i,  0 ,NZ+1);  
      Sigma_w(s2,i,NY,NZ) += W(i,NY+1,NZ+1);  
    }
    
    for(j=1;j<=NY;j++) {
      Sigma_w(s2,1 ,j,1 ) += W(  0 ,j,  0 );  
      Sigma_w(s2,NX,j,1 ) += W(NX+1,j,  0 );  
      Sigma_w(s2,1 ,j,NZ) += W(  0 ,j,NZ+1);  
      Sigma_w(s2,NX,j,NZ) += W(NX+1,j,NZ+1);  
    }
    
    for(k=1;k<=NZ;k++) {
      Sigma_w(s2,1 ,1 ,k) += W(  0 ,  0 ,k);  
      Sigma_w(s2,NX,1 ,k) += W(NX+1,  0 ,k);  
      Sigma_w(s2,1 ,NY,k) += W(  0 ,NY+1,k);  
      Sigma_w(s2,NX,NY,k) += W(NX+1,NY+1,k);  
    }
    
  } else { /* if manypoint */
    
    /* ------------------------------
     * Seven-point Laplacian formulas 
     * ------------------------------ */
    
    for(j=1;j<=NY;j++) {
      for(i=1;i<=NX;i++) {
	Sigma_w(s2,i,j,1 ) += W(i,j,0   );  
	Sigma_w(s2,i,j,NZ) += W(i,j,NZ+1); 
      }
    }
    
    for(k=1;k<=NZ;k++) {
      for(i=1;i<=NX;i++) {
	Sigma_w(s2,i,1 ,k) += W(i,0   ,k);  
	Sigma_w(s2,i,NY,k) += W(i,NY+1,k); 
      }
    }
    
    for(k=1;k<=NZ;k++) {
      for(j=1;j<=NY;j++) {
	Sigma_w(s2,1 ,j,k) += W(0   ,j,k); 
	Sigma_w(s2,NX,j,k) += W(NX+1,j,k); 
      }
    }
    
  } /* if manypoint else */
#undef W
#undef Sigma_w
} /* Impose_boundary_conditions_3D */
/* ========================================================================= */
